export type LoggerType = "Plugin" | "Coremod" | "API" | "Ignition" | "CommonModules" | "Components" | "Manager";
/**
 * Log a message to the console with a Replugged prefix.
 * @param type Type of the context of the message (e.g. API, Plugin, Coremod...)
 * @param name Name of the context of the message (e.g. Notices, SilentTyping, Badges...)
 * @param color Color of the prefix as hex or a CSS color
 * @param data Message(s) to log to the console
 */
export declare function log(type: LoggerType, name: string, color?: string, ...data: Parameters<typeof console.log>): void;
/**
 * Log a warning to the console with a Replugged prefix.
 * @param type Type of the context of the warning (e.g. API, Plugin, Coremod...)
 * @param name Name of the context of the warning (e.g. Notices, SilentTyping, Badges...)
 * @param color Color of the prefix as hex or a CSS color
 * @param data Thing(s) to print with the warning, same as the arguments would be for `console.warn`
 */
export declare function warn(type: LoggerType, name: string, color?: string, ...data: Parameters<typeof console.warn>): void;
/**
 * Log an error to the console with a Replugged prefix.
 * @param type Type of the context of the error (e.g. API, Plugin, Coremod...)
 * @param name Name of the context of the error (e.g. Notices, SilentTyping, Badges...)
 * @param color Color of the prefix as hex or a CSS color
 * @param data Thing(s) to print with the error, same as the arguments would be for `console.error`
 */
export declare function error(type: LoggerType, name: string, color?: string, ...data: Parameters<typeof console.error>): void;
/**
 * A convenient way to manage logging things to the console with colorful prefixes indicating their context.
 * Each `Logger` instance stores its context type, context name, and prefix color,
 * so you can use its {@link Logger.log log}, {@link Logger.warn warn}, and {@link Logger.error error}
 * methods in the same manner that you would use the `console` methods with the same names. The prefix
 * will be generated and prepended to the appropriate console message automatically.
 *
 * If you are only logging a single message with a prefix in your plugin, you may use {@link log}, {@link warn}, or {@link error}
 * instead of creating a `Logger`. Otherwise, using this class is much more convenient than specifying
 * the type, name, and color for every message.
 *
 * Example usage:
 * ```ts
 * import { Logger } from "replugged";
 *
 * const pluginLogger = Logger.plugin("SilentTyping");
 * pluginLogger.log("Hello", "world"); // Logs `[Replugged:Plugin:SilentTyping] Hello world`
 * ```
 */
export declare class Logger {
    type: LoggerType;
    name: string;
    color: string;
    /**
     *
     * @param type Type of the context (e.g. API, Plugin, Coremod...)
     * @param name Name of the context (e.g. Notices, SilentTyping, Badges...)
     * @param color Color of the prefix as hex or a CSS color
     */
    constructor(type: LoggerType, name: string, color?: string);
    /**
     * Logs a message to the console, with an identifying prefix managed by the Logger instance.
     * @param data Item(s) to print as a message
     * @remarks The arguments for this method are the same as the arguments for `console.log`.
     */
    log(...data: Parameters<typeof console.log>): void;
    /**
     * Logs a warning to the console, with an identifying prefix managed by the Logger instance.
     * @param data Item(s) to print as a warning
     * @remarks The arguments for this method are the same as the arguments for `console.warn`.
     */
    warn(...data: Parameters<typeof console.warn>): void;
    /**
     * Logs an error to the console, with an identifying prefix managed by the Logger instance.
     * @param data Item(s) to print as an error
     * @remarks The arguments for this method are the same as the arguments for `console.error`.
     */
    error(...data: Parameters<typeof console.error>): void;
    /**
     * Convenience method to create a new {@link Logger} for an API.
     * @internal
     * @param name Name of the API
     * @param color Color of the prefix as hex or a CSS color (default: blurple)
     * @returns {@link Logger} with type "API"
     */
    static api(name: string, color?: string): Logger;
    /**
     * Convenience method to create a new {@link Logger} for a coremod.
     * @internal
     * @param name Name of the coremod
     * @param color Color of the prefix as hex or a CSS color (default: blurple)
     * @returns {@link Logger} with type "Coremod"
     */
    static coremod(name: string, color?: string): Logger;
    /**
     * Convenience method to create a new {@link Logger} for a plugin.
     * @param name Name of the plugin
     * @param color Color of the prefix as hex or a CSS color (default: blurple)
     * @returns {@link Logger} with type "Plugin"
     */
    static plugin(name: string, color?: string): Logger;
}
