/// <reference types="react" />
import type { Fiber } from "react-reconciler";
import type { Jsonifiable } from "type-fest";
import { SettingsManager } from "./apis/settings";
/**
 * Loads a stylesheet into the document
 * @param path Path to the stylesheet
 * @returns Link element
 */
export declare const loadStyleSheet: (path: string) => HTMLLinkElement;
/**
 * Wait for an element to be added to the DOM
 * @param selector Element selector
 */
export declare function waitFor(selector: string): Promise<Element>;
/**
 * Async sleep function
 */
export declare function sleep(ms: number): Promise<void>;
/**
 * Get the React instance of an element
 * @param element Element to get the React instance of
 * @returns React instance
 * @throws If the React instance could not be found
 */
export declare function getReactInstance(element: Element): Fiber | null;
/**
 * Get the React owner instance of an element
 * @param element Element to get the React owner instance of
 * @returns React owner instance
 * @throws If the React owner instance could not be found
 */
export declare function getOwnerInstance(element: Element): React.Component & Record<string, unknown>;
/**
 * Force updates a rendered React component by its DOM selector
 * @param selector The DOM selector to force update
 * @param all Whether all elements matching that selector should be force updated
 */
export declare function forceUpdateElement(selector: string, all?: boolean): void;
/**
 * If the user is not in the server, join it. Otherwise, go to the server.
 * @param invite Invite code (eg "replugged")
 */
export declare function goToOrJoinServer(invite: string): Promise<void>;
export declare function openExternal(url: string): Promise<void>;
type ValType<T> = T | React.ChangeEvent<HTMLInputElement> | (Record<string, unknown> & {
    value?: T;
    checked?: T;
});
export declare function useSetting<T extends Record<string, Jsonifiable>, D extends keyof T, K extends Extract<keyof T, string>, F extends T[K] | undefined>(settings: SettingsManager<T, D>, key: K, fallback?: F): {
    value: K extends D ? NonNullable<T[K]> : F extends null | undefined ? T[K] | undefined : NonNullable<T[K]> | F;
    onChange: (newValue: ValType<T[K]>) => void;
};
export declare function useSettingArray<T extends Record<string, Jsonifiable>, D extends keyof T, K extends Extract<keyof T, string>, F extends T[K] | undefined>(settings: SettingsManager<T, D>, key: K, fallback?: F): [
    K extends D ? NonNullable<T[K]> : F extends null | undefined ? T[K] | undefined : NonNullable<T[K]> | F,
    (newValue: ValType<T[K]>) => void
];
type UnionToIntersection<U> = (U extends never ? never : (k: U) => void) extends (k: infer I) => void ? I & {
    all: () => I;
} : never;
type ObjectType = Record<never, never>;
type ExtractObjectType<O extends ObjectType[]> = O extends Array<infer T> ? UnionToIntersection<T> : never;
export declare function virtualMerge<O extends ObjectType[]>(...objects: O): ExtractObjectType<O>;
export type Tree = Record<string, unknown> | null;
type TreeFilter = string | ((tree: Tree) => boolean);
/**
 * All credit goes to rauenzi for writing up this implementation.
 * You can find the original source here:
 * <https://github.com/rauenzi/BDPluginLibrary/blob/683d6abc70f149a39e2f0433ffb65e55ca47c4e3/release/0PluginLibrary.plugin.js#L2585C15-L2611>
 *
 * @remarks Used mainly in findInReactTree
 */
export declare function findInTree(tree: Tree, searchFilter: TreeFilter, args?: {
    walkable?: string[];
    ignore?: string[];
    maxRecursion: number;
}): Tree | null | undefined;
/**
 * Find the component you are looking for in a tree, recursively.
 *
 * @param tree The tree to search through
 * @param searchFilter The filter. Either a string or a function. Should be unique
 * @param maxRecursion The max depth. Avoids call stack exceeded error.
 * @returns The component you are looking for
 */
export declare function findInReactTree(tree: Tree, searchFilter: TreeFilter, maxRecursion?: number): Tree | null | undefined;
export {};
